/* AICL 
version 1.1
author Fabrizio Riguzzi
*/


:-use_module(library(system)).
:-use_module(library(lists)).
%i due seguenti sono stati aggiunti per swi stand-alone
%:-use_module(library(edinburgh)).
%:-use_module(library(quintus)).
%:-set_prolog_flag(language,iso).
:-unknown(_S,fail).
:-dynamic setting/2.
:-dynamic mult/2.


/*:- set_prolog_flag(unknown,error).*/
:- set_prolog_flag(profiling,on).
:- set_prolog_flag(debug,on).
:- set_prolog_flag(discontiguous_warnings,on).
:- set_prolog_flag(single_var_warnings,on). 



setting(der_depth,20).
setting(beamsize,50).
setting(verbosity,3).
setting(significance_level,0).
setting(pruning,true).
/*allowed values 0.995 / 0.99 / 0.975 / 0.95 / 0.90 / 0.75 / 0.0 */
setting(min_coverage,3).
setting(min_accuracy,0.75).
setting(max_nodes,100).
setting(heur,acc).
/* allowed values: acc, laplace */
setting(optimal,yes).
/* allowed values: yes, no */
setting(max_length,8).

:-consult(ref_op).

i(File):-
    generate_file_names(File,FileKB,FileBG,FileOut,FileL),
    statistics(runtime,[_,_]),
    load_bg(FileBG),
    load_models(FileKB,HB,ModulesList),
%   load_files(FileL,[silent(true)]), %SWI
    reconsult(FileL), %yap
%    divide_pos_neg(ModulesList,[],Neg,[],Pos), %neg al posto dei pos
    divide_pos_neg(ModulesList,[],Pos,[],Neg), %nomrmale
    open(FileOut,write,Stream),
    format("/* Inizio l'apprendimento da ~p */~n",[File]),
    induce(Pos,Neg,NegRem,N,ProgramRed,Stream),
    remove_red(Pos,ProgramRed,[],Program),
    statistics(runtime,[_,T]),
    T1 is T /1000,
    findall(setting(A,B),setting(A,B),L),
    length(NegRem,NR),
    %findall(template(HeadType,BodyType,Name,Head,Body),template(HeadType,BodyType,Name,Head,Body),LT),
    %numbervars(LT,0,_),
    setting(optimal,Opt),
    set_output(Stream),
    format("/* Learning S time ~f seconds. Generated rules */~N",[T1]),
    format("/* Number of rules ~d */~n",[N]),
    format("/* ~p */~n~n",[L]), 
    format("/* Negative examples remaining: ~d~n~p~n*/~n",[NR,NegRem]),   
    format("/* Language bias ~n~p~n~p~n*/~n",[optimal(Opt),LT]),
%    print_list(Program),
    set_output(user_output),
    format("/* Learning time ~f seconds. Generated rules */~N",[T1]),
    format("/* Number of rules ~d */~n",[N]),
    format("/* ~p */~n~n",[L]),    
    format("/* Negative examples remaining: ~d~n~p~n*/~n",[NR,NegRem]),   
    print_list(Program),
    close(Stream),
		clear_kb(HB).


divide_pos_neg([],Pos,Pos,Neg,Neg):-!.
    
divide_pos_neg([MH|MT],PosIn,PosOut,NegIn,NegOut):-
    (pos(MH)->
        PosOut=[MH|Pos],
        NegOut=Neg
    ;
        PosOut=Pos,
        NegOut=[MH|Neg]
    ),
    divide_pos_neg(MT,PosIn,Pos,NegIn,Neg).
        

induce(Pos,Neg,NegRem,NR,Program,S):-
    prior_prob(Pos,Neg,NP,NN),
    covering_loop(Pos,Neg,NegRem,NP,NN,0,NR,[],Program,S).

prior_prob(Pos,Neg,NP,NN):-
    total_number(Pos,0,NP),
    total_number(Neg,0,NN),
    assert(npt(NP)),
    assert(nnt(NN)).
    
total_number([],N,N):-!.

total_number([H|T],NIn,NOut):-
  (mult(H,M)->
    N1 is NIn+M
  ;
    N1 is NIn+1
  ),
  total_number(T,N1,NOut).

covering_loop(_Pos,[],[],Rules,Rules,_S):-!.

/* some eminus still to cover: generate a new clause */
covering_loop(Eplus,Eminus,EminusRem,NP,NN,NR,NR2,Rulesin,Rulesout,S):-
        print_ex_rem(Eplus,Eminus),
        initialize_agenda(Eplus,Eminus,NP,NN,Agenda,BestClause),
        specialize(Agenda,Eplus,Eminus,NP,NN,0,BestClause,(Name,BestClauseOut,Heur,(NC,PC,Emc,Epnc))),
        (BestClauseOut=null->
            format("No more significant clauses.~n~n",[]),
            print_ex_rem(Eplus,Eminus),
            Rulesout=Rulesin,
            NR2=NR,
            EminusRem=Eminus
        ;
            set_output(S),
            write_clause(BestClauseOut),
            NR1 is NR+1,	    
            %MODIFICATO
	    %numbervars(Name,0,_,[functor_name(xarg)]),
            numbervars(Name,0,_),
            format("/* Rule n. ~d ",[NR1]), 
            write_term(Name,[numbervars(true)]),
            format(" ~p ~p ~p ~n",[acc(Heur), negcov(NC), poscov(PC)]),
            format("Neg traces ruled out:#~p */~n~n~n",[Emc]),
            %format("/* Rule n. ~d ~p ~p ~p ~p */~n",[NR1,Name,acc(Heur),negcov(NC),poscov(PC)]),
            %test_body(BestClauseOut,Eplus,NBODY,S),
            %total_number(NBODY,0,NB),
            %format("/* Positivi ~p */~n~n",[NB]),
            set_output(user_output),
            print_new_clause(Name,BestClauseOut,Heur,NC,PC,Emc,Epnc),
            flush_output(S),
            remove_cov_examples(Emc,Eminus,EminusOut),
            length(EminusOut,NN1),
            Rulesout=[rule(Name,BestClauseOut,(heur(Heur),negcov(NC),poscov(PC),emc(Emc),epnc(Epnc)))|Rules1],
            covering_loop(Eplus,EminusOut,EminusRem,NP,NN1,NR1,NR2,Rulesin,Rules1,S)
        ).


remove_cov_examples([],Eminus,Eminus):-!.
    
remove_cov_examples([Ex|Rest],Eminus,EminusOut):-
    delete(Eminus,Ex,Eminus1),
    remove_cov_examples(Rest,Eminus1,EminusOut).



/* specialize(Agenda,Clout,AllowedLiterals,Eplus,Eminus,
        Deltain) */

/* stopping criterion (1): empty agenda */

specialize([],_Ep,_Em,_NPT,_NNT,_N,BestClause,BestClause):-!.

specialize(_Ag,_Ep,_Em,_NPT,_NNT,N,BestClause,BestClause):-
  setting(max_nodes,NMax),
  N>NMax,!.

specialize(_Ag,_Ep,_Em,_NPT,_NNT,_N,(Name,BestClause,H,(NN,NP,Emc,Epnc)),(Name,BestClause,H,(NN,NP,Emc,Epnc))):-
  H=1.0,
  setting(min_coverage,MC),
  NN>=MC,!.

specialize(Agenda,Ep,Em,NPT,NNT,N,BCIn,BCOut):-
    generate_new_agenda(Ep,Em,NPT,NNT,Agenda,[],NewAgenda,BCIn,BC1),
    N1 is N+1,!,
    specialize(NewAgenda,Ep,Em,NPT,NNT,N1,BC1,BCOut).
    
generate_new_agenda(_Ep,_Em,_NPT,_NNT,[],NewAg,NewAg,BC,BC):-!.    

generate_new_agenda(Ep,Em,NPT,NNT,[(Name,Node,_Heur,_NN)|Rest],
        NAgIn,NAgOut,BCIn,BCOut):-
    findall(NewNode,refine(Node, NewNode),Ref), 
    evaluate_all_refinements(Ep,Em,NPT,NNT,Ref,Name,NAgIn,NAg1,BCIn,BC1),!,
    generate_new_agenda(Ep,Em,NPT,NNT,Rest,NAg1,NAgOut,BC1,BCOut).
        
evaluate_all_refinements(_Ep,_Em,_NPT,_NNT,[],_Name,NAg,NAg,BC,BC):-!.

evaluate_all_refinements(Ep,Em,NPT,NNT,[HRef|TRef],Name,NAgIn,NAgOut,(NameIn,BCIn,HeurIn,DetIn),
        (NameOut,BCOut,HeurOut,DetOut)):-
    generate_query(HRef,Query,VI),
    test_clause_pos(Ep,Query,VI,0,NP,[],Epc),
    test_clause_neg(Em,Query,VI,0,NN,[],Emc),
    deleteall(Ep,Epc,Epnc),
%    length(Em,NNTC),
    setting(min_coverage,MC),
    setting(min_accuracy,MA),
    (setting(heur,laplace)->
        Heuristic is (NN+1)/(NPT-NP+NN+2)
    ;
      Den is (NPT-NP+NN),
      (Den=0->
        Heuristic is 0     
      ;         
        Heuristic is (NN)/Den      
      )
    ),
    ((Heuristic > HeurIn,Heuristic>MA
%    ,statistically_significant(NP,NN,NNTC,NPT,NNT)
    ,NN>=MC)->
        Name1=Name,
        BC1 = HRef,
        Heur1 = Heuristic,
        Det1 = (NN,NP,Emc,Epnc)
    ;
        Name1=NameIn,
        BC1=BCIn,
        Heur1 = HeurIn,
        Det1 = DetIn
    ),
    (prune(NN,Heuristic,HeurIn)->
        NAg1=NAgIn
    ;
        setting(beamsize,BS),
        print_ref(Name,HRef,Heuristic,NN,NP,Emc,Epnc),
        insert_in_order((Name,HRef,Heuristic,NN),BS,NAgIn,NAg1)
    ),!,
    evaluate_all_refinements(Ep,Em,NPT,NNT,TRef,Name,NAg1,NAgOut,(Name1,BC1,Heur1,Det1),
        (NameOut,BCOut,HeurOut,DetOut)).


test_body(((H,_HL):-(B,_BL)),Ep,NBODY,S):-
	generate_query((([],[]):-(B,_BL)),Query,VI),
	set_output(S),
	total_number(Ep,0,Nep),
	format("/* Query ~d ~p ~p */~n~n",[Nep,Query,VI]),
	test_clause_pos(Ep,Query,VI,0,NP,[],NBODY).


generate_query(((H,_HL):-(B,_BL)),QA,VI):-
  process_head(H,HA,VI),
  add_int_atom(B,B1,VI),
  append(B1,HA,Q),
  list2and(Q,QA).

process_head([],[],_VI).  

process_head([(+,D,_DL)|T],[\+(DA)|T1],VI):-
  add_int_atom(D,D1,VI),
  list2and(D1,DA),
  process_head(T,T1,VI).
  
process_head([(+=,D,_DL)|T],[\+(DA)|T1],VI):-
  add_int_atom(D,D1,VI),
  list2and(D1,DA),
  process_head(T,T1,VI).

process_head([(-,D,_DL)|T],[\+(\+(DA))|T1],VI):-
  add_int_atom(D,D1,VI),
  list2and(D1,DA),
  process_head(T,T1,VI).
  
process_head([(-=,D,_DL)|T],[\+(\+(DA))|T1],VI):-
  add_int_atom(D,D1,VI),
  list2and(D1,DA),
  process_head(T,T1,VI).

add_int_atom([],[],_VI).

add_int_atom([H|T],[H|T1],VI):-
  builtin(H),!,
  add_int_atom(T,T1,VI).

add_int_atom([H|T],[H1|T1],VI):-
  H=..[F|Args],
  H1=..[F,VI|Args],
  add_int_atom(T,T1,VI).



    
prune(_NPT,_NNT,NN,_H,HeurBestClause):-
    setting(pruning,true),
    BestHeur is (NN+1)/(NN+2),
    BestHeur<HeurBestClause.
    
prune(NPT,NNT,NN,_H,_HeurBestClause):-
    setting(pruning,true),
    BestLR is -2*NN*log10(NNT/(NNT+NPT)),
    setting(significance_level,X),
    sig_threshold(X,T),
    BestLR<T.
    
statistically_significant(NP,NN,_NNTC,NPT,NNT):-
    PMinusCMinus is NN/(NN+NPT-NP),
    (PMinusCMinus=:=0->
        LikelihoodRatio is 2*(NN+NPT-NP)*log10(1/(NPT/(NPT+NNT)))
    ;
        (PMinusCMinus=:=1->
            LikelihoodRatio is 2*(NN+NPT-NP)*log10(1/(NNT/(NPT+NNT)))
        ;
            LikelihoodRatio is 2*(NN+NPT-NP)*
                (PMinusCMinus*log10(PMinusCMinus/(NNT/(NPT+NNT))) +
                (1-PMinusCMinus)*log10((1-PMinusCMinus)/(NPT/(NPT+NNT))))
        )
    ),
    setting(significance_level,X),
    sig_threshold(X,T),
    LikelihoodRatio>T.

statistically_significant_pos(NP,NN,NNTC,NPT,NNT):-
    PPlusC is NP/(NP+NNTC-NN),
    (PPlusC=:=0->
        LikelihoodRatio is 2*(NP+NNTC-NN)*log10(1/(NNT/(NPT+NNT)))/log10(2)
    ;
        (PPlusC=:=1->
            LikelihoodRatio is 2*(NP+NNTC-NN)*log10(1/(NPT/(NPT+NNT)))/log10(2)
        ;
            LikelihoodRatio is 2*(NP+NNTC-NN)*
                (PPlusC*log10(PPlusC/(NPT/(NPT+NNT)))/log10(2) +
                (1-PPlusC)*log10((1-PPlusC)/(NNT/(NPT+NNT)))/log10(2))
        )
    ),
    setting(significance_level,X),
    sig_threshold(X,T),
    LikelihoodRatio>T.

sig_threshold(0.995,7.88).    
sig_threshold(0.99,6.64).
sig_threshold(0.975,5.02).
sig_threshold(0.95,3.84).
sig_threshold(0.90,2.71).
sig_threshold(0.75,1.32).
sig_threshold(0,0).

list2andHead([],false):-!.

list2andHead(HeadList,Head):-
    list2and(HeadList,Head).

list2andBody([],true):-!.

list2andBody(BodyList,Body):-
    list2and(BodyList,Body).
    

initialize_agenda(Pos,Neg,NP,NN,L,(NameOut,BCOut,HeurOut,DetOut)):-
  findall( (Name,[],H,B,[]),template(dynamic,fixed,Name,H,B),L_dynamic_fixed1),
  evaluate_fixed(Pos,Neg,NP,NN,L_dynamic_fixed1,[],L_dynamic_fixed,(null,null,0,(0,_,_,_)),(NameOut1,BCOut1,HeurOut1,DetOut1)),
  findall( (Name,H,[],B,[]),template(fixed,fixed,Name,H,B),L_fixed_fixed1),
  evaluate_fixed(Pos,Neg,NP,NN,L_fixed_fixed1,L_dynamic_fixed,L_fixed_fixed,(NameOut1,BCOut1,HeurOut1,DetOut1),(NameOut2,BCOut2,HeurOut2,DetOut2)),
  findall( (Name,H,[],[],B),template(fixed,dynamic,Name,H,B),L_fixed_dynamic1),
  evaluate_fixed(Pos,Neg,NP,NN,L_fixed_dynamic1,L_fixed_fixed,L_fixed_dynamic,(NameOut2,BCOut2,HeurOut2,DetOut2),(NameOut,BCOut,HeurOut,DetOut)),
  findall( (Name,(([],H):-([],B)),0,NN ),template(dynamic,dynamic,Name,H,B),L_dynamic_dynamic),
  append(L_fixed_dynamic,L_dynamic_dynamic,L).

evaluate_fixed(_Pos,_Neg,_NPT,_NNT,[],Ag,Ag,BC,BC).
 
evaluate_fixed(Ep,Em,NPT,NNT,[(Name,H,HL,B,BL)|TRef],NAgIn,NAgOut,(NameIn,BCIn,HeurIn,DetIn),
        (NameOut,BCOut,HeurOut,DetOut)):-
    extract_disj(H,H1),
    generate_query(((H1,HL):-(B,BL)),Query,VI),
    test_clause_pos(Ep,Query,VI,0,NP,[],Epc),
    test_clause_neg(Em,Query,VI,0,NN,[],Emc),
    deleteall(Ep,Epc,Epnc),
%    length(Em,NNTC),
    setting(min_coverage,MC),
    setting(min_accuracy,MA),
    (setting(heur,laplace)->
        Heuristic is (NN+1)/(NPT-NP+NN+2)
    ;
      Den is (NPT-NP+NN),
      (Den=0->
        Heuristic is 0     
      ;         
        Heuristic is (NN)/Den      
      )
    ),
    DetIn=(NNIn,_,_,_),
    (((Heuristic > HeurIn;Heuristic=:=HeurIn,NN>NNIn),Heuristic>MA
%    ,statistically_significant(NP,NN,NNTC,NPT,NNT)
    ,NN>=MC)->
        Name1=Name,
        BC1 = ((H1,HL):-(B,BL)),
        Heur1 = Heuristic,
        Det1 = (NN,NP,Emc,Epnc)
    ;
        Name1=NameIn,
        BC1=BCIn,
        Heur1 = HeurIn,
        Det1 = DetIn
    ),
    print_ref(Name,((H1,HL):-(B,BL)),Heuristic,NN,NP,Emc,Epnc),
    (prune(NPT,NNT,NN,Heuristic,HeurIn)->
        NAg1=NAgIn
    ;
        insert_in_order((Name,((H1,HL):-(B,BL)),Heuristic,NN),10000000,NAgIn,NAg1)
    ),!,
    evaluate_fixed(Ep,Em,NPT,NNT,TRef,NAg1,NAgOut,(Name1,BC1,Heur1,Det1),
        (NameOut,BCOut,HeurOut,DetOut)).


extract_disj([],[]).

extract_disj([(S,D)|T],[(S,D,[])|T1]):-
	extract_disj(T,T1).  
/*
initialize_clause(NAME,[],H,B,[]):-
	template(fixedbody,NAME,H,B).
	

initialize_clause(NAME,[],H,[],B):-
	template(dynamicbody,NAME,H,B).
*/
	
	
  
  
  

gen_cov_eminus([],[]):-!.

gen_cov_eminus([H|T],[(H,[])|T1]):-  
    gen_cov_eminus(T,T1).
    
print_ex_rem(Eplus,Eminus):-
        setting(verbosity,V),
        V>0,
        length(Eplus,Lp),
        format("Positive examples remaining: ~d~N~p~N~N",[Lp,Eplus]),
        length(Eminus,Lm),
        format("Negative examples remaining: ~d~N~p~N~N",[Lm,Eminus]).

insert_in_order(C,BeamSize,[],[C]):-
        BeamSize>0,!.

insert_in_order(_NewClauseItem,0,Beam,Beam):-!.


insert_in_order((Name,HRef,Heuristic,NN),
        BeamSize,
        [(Name1,HRef1,Heuristic1,NN1)|RestBeamIn],
        BeamOut):-
    (Heuristic>Heuristic1;(Heuristic=Heuristic1,NN>NN1)),!,
    % bigger heuristic, insert here
    NewBeam=[(Name,HRef,Heuristic,NN),(Name1,HRef1,Heuristic1,NN1)|RestBeamIn],
    length(NewBeam,L),
    (L>BeamSize->
        nth(L,NewBeam,_Last,BeamOut)
        
    ;
        BeamOut=NewBeam
    ).

insert_in_order((Name,HRef,Heuristic,NN),BeamSize,
        [(Name1,HRef1,Heuristic1,NN1)|RestBeamIn],
        [(Name1,HRef1,Heuristic1,NN1)|RestBeamOut]):-
    BeamSize1 is BeamSize -1,
    insert_in_order((Name,HRef,Heuristic,NN),BeamSize1,RestBeamIn,
                RestBeamOut).


        



/* test_clause_pos(PosEx,(Head:-Body),NIn,NOut,CovIn,CovOut) returns in NOut
the number of ex where the clause is true and in CovOut a list of covered examples */    
test_clause_pos([],_Q,_VI,N,N,Ec,Ec):-!.

test_clause_pos([Module|Rest],Q,VI,NIn,NOut,EcIn,EcOut):-
  copy_term(r(Q,VI),r(Q1,VI1)),
  VI1=Module,
    (call(Q1)->
        N is NIn,
        Ec=EcIn
    ;
      (mult(Module,M)->
        N is NIn+M
      ;
        N is NIn + 1
      ),
        Ec =[Module|EcIn]
    ),
    test_clause_pos(Rest,Q,VI,N,NOut,Ec,EcOut).                

test_clause_neg([],_Q,_VI,N,N,Ec,Ec):-!.

test_clause_neg([Module|Rest],Q,VI,NIn,NOut,EcIn,EcOut):-
  copy_term(r(Q,VI),r(Q1,VI1)),
  VI1=Module,
    (call(Q1)->
      (mult(Module,M)->
        N is NIn+M
      ;
        N is NIn + 1
      ),
        Ec =[Module|EcIn]
    ;
        N is NIn,
        Ec=EcIn
    ),
    test_clause_neg(Rest,Q,VI,N,NOut,Ec,EcOut).                

distribute_not(L,\+ L):-
    L\=(_,_),!.

distribute_not((L,RestL),(\+ L ,NewRestL)):-
    distribute_not(RestL,NewRestL).

remove_red(_Pos,[],P,P).

remove_red(Pos,[rule(Name,C,Stat)|T],PIn,POut):-
  reduce_clause(Pos,C,CRed),
  append(PIn,[rule(Name,CRed,Stat)],P1),
  remove_red(Pos,T,P1,POut).  

reduce_clause(Pos,((H,HL):-(B,BL)),((HR,HL):-(B,BL))):-
  reduce_head(B,Pos,H,[],HR).

reduce_head(_B,_Pos,[],Head,Head).
  
reduce_head(B,Pos,[H|T],HeadIn,HeadOut):-
  generate_query((([H],_):-(B,_)),Q,VI),
  test_clause_pos(Pos,Q,VI,0,NP,[],Epc),
  (NP=0->
    Head1=HeadIn,
    Pos1=Pos
  ;
    append(HeadIn,[H],Head1),
    deleteall(Pos,Epc,Pos1)
  ),
  reduce_head(B,Pos1,T,Head1,HeadOut).


deleteall(L,[],L).

deleteall(L,[H|T],LOut):-
  delete(L,H,L1),
  deleteall(L1,T,LOut).


load_models(File,HB,ModulesList):-
    open(File,read,Stream),
    read_models(Stream,[],HB,ModulesList),
    close(Stream).

read_models(Stream,HB0,HB,[Name1|Names]):-
    read(Stream,begin(model(Name))),!,
    (number(Name)->
        name(Name,NameStr),
        append("i",NameStr,Name1Str),
        name(Name1,Name1Str)
    ;
        Name1=Name
    ),
    read_all_atoms(Stream,HB0,HB1,Name1),
    read_models(Stream,HB1,HB,Names).

read_models(_S,HB,HB,[]).

read_all_atoms(Stream,HB0,HB,Name):-
    read(Stream,Atom),
    Atom \=end(model(_Name)),!,
    Atom=..[Pred|Args],
    Atom1=..[Pred,Name|Args],
    assertz(Atom1),
    functor(Atom1,F,A),
    (member(F/A,HB0)->
    	HB1=HB0
    ;
    	HB1=[F/A|HB0]
    ),
    read_all_atoms(Stream,HB1,HB,Name).    

    
read_all_atoms(_S,HB,HB,_N).


title(File,Stream):-
        max_spec_steps(Spec),
        der_depth(Der),
        beamsize(B),
        ver(V),
        setting(verbosity,Ver),
        verapp(Vapp),
        min_cov(MC),
        format(Stream,"~N ~N/*~NACL1 ver ~a AbdProofProc. ver ~a~NFile name: ~a~N",
                [V,Vapp,File]),
        format(Stream,"Max spec steps=~d, Beamsize=~d,~NDerivation depth=~d, Verbosity=~d, Minimum coverage=~d~N*/~N",
                        [Spec,B,Der,Ver,MC]).
  

list2and([],true):-!.

list2and([X],X):-!.

list2and([H|T],(H,Ta)):-
        list2and(T,Ta).

and2list(true,[]):-!.


and2list((H,Ta),[H|T]):-!,
        and2list(Ta,T).

and2list(X,[X]).

print_list([]):-!.

print_list([rule(Name,C,Stat)|Rest]):-
    numbervars(C,0,_M),
    write_clause(C),
    format("/* ~p ~p */~n~n",[Name,Stat]),    
    print_list(Rest).

% CODICE PER SWI
%load_bg(FileBG):-
%  (exists_file(FileBG)->
%    open(FileBG,read,S), 
%    read_all_atoms_bg(S),
%    close(S)
%  ;
%    true
%  ).  

% CODICE PER YAP
load_bg(FileBG):-
  (file_exists(FileBG)->
    open(FileBG,read,S), 
    read_all_atoms_bg(S),
    close(S)
  ;
    true
  ). 

%prune(NN,_H,HeurBestClause):-
%    setting(pruning,true),
%    BestHeur is (NN+1)/(NN+2),
%    BestHeur<HeurBestClause.
    
%prune(NN,_H,_HeurBestClause):-
%    setting(pruning,true),
%    nnt(NNT),
%    npt(NPT),
%    BestLR is -2*NN*log(10,NNT/(NNT+NPT)),
%    setting(significance_level,X),
%    sig_threshold(X,T),
%    BestLR<T.

 

read_all_atoms_bg(Stream):-
    read(Stream,Atom),
    (Atom=end_of_file->
      true
    ;
      process(Atom,Atom1),
      assertz(Atom1),
      read_all_atoms_bg(Stream)
    ).    

process((H:-B),(H1:-B1)):-!,
  add_int_atom([H],[H1],VI),
  and2list(B,BL),
  add_int_atom(BL,BL1,VI),
  list2and(BL1,B1).  
      
process(H,H1):-!,
  add_int_atom([H],[H1],_VI).

print_new_clause(Name,C,Heur,NC,PC,_Emc,_Epnc):-
        setting(verbosity,V),
        V>0,
        format(" ~N ~NGenerated clause:~n",[]),
        write_clause(C),
        nl,
        copy_term(Name,Name1),
        numbervars(Name1,0,_),
        format("Name:~p~n",[Name1]),
        format("Heuristic:~p~n",[Heur]),
        format("Neg ex ruled out:#~p~n",[NC]),
%        format("Neg ex ruled out:#~p~n",[Emc]),
        format("Covered pos ex:#~p~n",[PC]),
%        format("Covered pos ex:#~p~n",[Epnc]),
%%        format("correct: ~a, Np=~d, Npa=~d, Nm=~d, Nma=~d\c
%                ~NPos ex cov: ~p~NNeg ex cov: ~p~NAbduced literals: ~p~N ~N",
%                [C,Np,Npa,Nm,Nma,
%                Epluscovered,Eminuscovered,NewDelta]),
        (V>3->
                get0(_)
        ;
                true
        ).

write_clause(((H,_HL):-(B,_BL))):-
  copy_term(c(H,B),c(H1,B1)),
  numbervars((H1,B1),0,_M),
    write('\t'),
    (B1=[]->
      write(true)
    ;
      write_list(B1)
    ),
    nl,
    write('--->'),
    nl,
    write_head(H1).

write_head([]):-
  write('\t'),
  write('false.'),nl.

write_head([(Sign,[A|T],_DL)]):-!,
   write('\t'),
  (Sign = '+' ->
    write('E(')
  ; (Sign = '+=' ->
  	write('E(')
    ;
  	write('EN(')
    ) 		
  ),
  	%MODIFICA
    %  write(A),
	write_term(A,[numbervars(true)]),
  (T=[]->
    write(')')
  ;
    write(')\n\t/\\'),
    write_list(T)
  ),
  write('.'),
  nl.

write_head([(Sign,[A|T],_DL)|HT]):-!,
   write('\t'),
  (Sign = '+' ->
    write('E(')
  ; (Sign = '+=' ->
  	write('E(')
    ;
  	write('EN(')
    ) 		
  ),
  	%MODIFICA
    %  write(A),
	write_term(A,[numbervars(true)]),
  (T=[]->
    write(')')
  ;
    write(')\n\t/\\'),
    write_list(T)
  ),
  nl,
  write('\\/'),nl,
  write_head(HT).


/*
write_head([(Sign,[h(Ev,Time)|T],_DL)]):-!,
   write('\t'),
  (Sign = '+' ->
    write('E(')
  ;
    write('EN(')
  ),
  write(Ev),
  write(','),
  write(Time),
  write(')\n\t/\\'),
  write_list(T),
  nl.

write_head([(Sign,[h(Ev,Time)|T],_DL)|HT]):-!,
   write('\t'),
  (Sign= '+' ->
    write('E(')
  ;
    write('EN(')
  ),
	%MODIFICA
    %  write(Ev),
	write_term(Ev,[numbervars(true)]),
  write(','),
  write(Time),
  write(')\n\t/\\'),
  write_list(T),nl,
  write('\\/'),nl,
  write_head(HT).
*/

write_list([H]):-!,
  (H=h(E,Time)->
    write('H('),
	%MODIFICA
    %write(E),
	write_term(E,[numbervars(true)]),
    write(','),
    write(Time),
    write(')')
  ;
	%MODIFICA
    %write(H)
	write_term(H,[numbervars(true)])
  ).

write_list([H|T]):-
  (H=h(E,Time)->
    write('H('),
	%MODIFICA
    %write(E),
	write_term(E,[numbervars(true)]),
    write(','),
    write(Time),
    write(')')
  ;
	%MODIFICA
    %write(H)
	write_term(H,[numbervars(true)])
  ),
  write('\n\t/\\'),
  write_list(T).

print_ref(Name,C,Heur,NC,PC,_Emc,_Epnc):-
        setting(verbosity,V),
        V>0,
        format("Refinemnt:~n",[]),
        write_clause(C),
        copy_term(Name,Name1),
        numbervars(Name1,0,_),
        format("Name:~p~n",[Name1]),
        format("Heuristic:~p~n",[Heur]),
        format("Neg ex ruled out:#~p~n",[NC]),
        format("Covered pos ex:#~p~n",[PC]),
      nl,
        (V>3->
                get0(_)
        ;
                true
        ).


generate_file_names(File,FileKB,FileBG,FileOut,FileL):-
        name(File,FileString),
        append(FileString,".kb",FileStringKB),
        name(FileKB,FileStringKB),
        append(FileString,".bg",FileStringBG),
        name(FileBG,FileStringBG),
        append(FileString,".l",FileStringL),
        name(FileL,FileStringL),
        append(FileString,".icl.out",FileOutString),
        name(FileOut,FileOutString).

clear_kb([]).

clear_kb([F/A|T]):-
	abolish(F,A),
	clear_kb(T).
        
builtin(_A is _B).
builtin(_A > _B).
builtin(_A < _B).
builtin(_A >= _B).
builtin(_A =< _B).
builtin(_A =:= _B).
builtin(_A =\= _B).
builtin(true).
builtin(false).
builtin(_A = _B).
builtin(_A==_B).
builtin(_A\=_B).
builtin(_A\==_B).
builtin(var(_A)).
builtin(nonvar(_A)).
builtin(!).

set(Parameter,Value):-
  retract(setting(Parameter,_)),
  assert(setting(Parameter,Value)).
  
/*
portray(xarg(N)) :-
    format('X~w',[N]).
*/
