% refinement operator for bodies
%
%
% Se non scelgo i raffinamento ottimale o raffino il body o la testa
% Head la testa attuale
% HeadList la testa presa a partire dal template
% Body il body attuale
% BodyList il body preso dal template
%
refine(((H,HL):-(B,BL)),((H1,HL1):-(B1,BL1))):-
  length(H,HN),
  length(B,BN),
  N is HN+BN,
  setting(max_length,ML),
  N=<ML,
  (setting(optimal,no)->
    ((refine_body_no(B,BL,B1,BL1),H1=H,HL1=HL)
    ;
      (refine_head_no(H,HL,H1,HL1),B1=B,BL1=BL)
     )  
  ;
    refine(B,BL,B1,BL1,H,HL,H1,HL1)
  ).

% raffino il body aggiungendo uno dei possibili 
refine_body_no(B,BL,NewB,NewBL):-
  member(E,BL),
  delete(E,BL,NewBL),
%  \+ member_eq(E,B),
  append(B,[E],NewB).

% posso raffinare il body
refine(B,BL,B1,BL1,H,HL,H,HL):-
  refine_body(B,BL,B1,BL1).

% se raffino la testa non posso pi raffinare il body quindi metto BL a []
refine(B,_BL,B,[],H,HL,H1,HL1):-
  refine_head(H,HL,H1,HL1).

% raffino il body aggiungendo un elemento e quindi riducendo la BL
refine_body(B,[H|T],NewB,T):-
  append(B,[H],NewB).

% posso raffinare il body anche non aggiungendo nulla
refine_body(B,[_H|T],NewB,BL):-
  refine_body(B,T,NewB,BL).

% Raffinamento della testa aggiungendo un disjoint
% [(+,[HD|TD],TD)] significa che per gli E inizio aggiungendo tutti i vincoli e  mi segno in TD quali sono cos li posso eliminare
% [(+,[HD|TD],TD)] significa che per gli EN inizio mettendo solo l'EN e mi segno in TD quali sono i vincoli da aggiungere
%
% Originale
%refine_head_no(H,HL,NewH,HL):-
%  member(HH,HL),
%  (HH=(+,[HD|TD])->
%    append(H,[(+,[HD|TD],TD)],NewH)
%  ;
%    HH=(-,[HD|TD]),
%    append(H,[(-,[HD],TD)],NewH)
%  ).

refine_head_no(H,HL,NewH,NewHL):-
  member(HH,HL),
  delete(HH,HL,NewHL),
  (HH=(+,[HD|TD])->
    append(H,[(+,[HD|TD],TD)],NewH)
  ;
    (HH=(-,[HD|TD])->
    	append(H,[(-,[HD],TD)],NewH)
    ;
    	(HH=(+=,[HD|TD])->
    		append(H,[(+=,[HD|TD],[])],NewH)
    	;
    		HH=(-=,[HD|TD]),
    		append(H,[(-=,[HD|TD],[])],NewH)
    	)	
    )	
  ).

% Raffinamento della testa, raffinando un disjoint
refine_head_no(H,HL,NewH,HL):-
  refine_disj(H,NewH).



refine_head(H,HL,H1,HL1):-
  add_disj(H,HL,H1,HL1).

refine_head(H,_HL,NewH,[]):-
  refine_disj(H,NewH).
  
% Originale  
%add_disj(H,[HH|T],NewH,T):-
%  (HH=(+,[HD|TD])->
%    append(H,[(+,[HD|TD],TD)],NewH)
%  ;
%    HH=(-,[HD|TD]),
%    append(H,[(-,[HD],TD)],NewH)
%  ).

add_disj(H,[HH|T],NewH,T):-
  (HH=(+,[HD|TD])->
    append(H,[(+,[HD|TD],TD)],NewH)
  ;
    (HH=(-,[HD|TD])->
    	append(H,[(-,[HD],TD)],NewH)
    ;
    	(HH=(+=,[HD|TD])->
    		append(H,[(+=,[HD|TD],[])],NewH)
    	;
    		HH=(-=,[HD|TD]),
    		append(H,[(-=,[HD|TD],[])],NewH)
    	)	
    )	
  ).
  
  

add_disj(H,[_HH|T],NewH,HL):-
  add_disj(H,T,NewH,HL).


% Raffinamento del disjoint nella testa 
%
refine_disj([(Sign,D,DL)|T],[(Sign,D1,DL1)|T]):-
  (setting(optimal,no)->
    refine_single_disj_no(Sign,D,DL,D1,DL1)
  ;
    refine_single_disj(Sign,D,DL,D1,DL1)
  ).

% Raffinamento di un disjoint interno
refine_disj([D|T],[D|T1]):-
  refine_disj(T,T1).


% Raffino una E togliendo un vincolo
refine_single_disj_no(+,D,DL,D1,DL):-
  member(E,D),
  delete(D,E,D1).

% Raffino un EN agiungendo un vincolo
refine_single_disj_no(-,D,DL,D1,DL1):-
  member(E,DL),
  delete(E,DL,DL1),
%  \+ member_eq(E,D),
  append(D,[E],D1).
  
% Gli elementi con += vanno lasciati intonsi
%refine_single_disj_no(+=,D,DL,D,DL). 

% Gli elementi con -= vanno lasciati intonsi
%refine_single_disj_no(-=,D,DL,D,DL). 


refine_single_disj(+,D,[H|T],D1,T):-
  delete(D,H,D1).

refine_single_disj(+,D,[_H|T],D1,DL1):-
  refine_single_disj(+,D,T,D1,DL1).

refine_single_disj(-,D,[H|T],D1,T):-
  append(D,[H],D1).

refine_single_disj(-,D,[_H|T],D1,DL1):-
  refine_single_disj(-,D,T,D1,DL1).

% Gli elementi con += vanno lasciati intonsi
%refine_single_disj(+=,D,DL,D,DL).

% Gli elementi con -= vanno lasciati intonsi
%refine_single_disj(-=,D,DL,D,DL).



number(+inf,Inf):-
    Inf is inf, !.
number(-inf,MInf):-
    MInf is -inf, !.
number(X,Y):-
    Y is X, !.



%--------------
aleph_member1(H,[H|_]):- !.
aleph_member1(H,[_|T]):-
    aleph_member1(H,T).

aleph_member2(X,[Y|_]):- X == Y, !.
aleph_member2(X,[_|T]):-
    aleph_member2(X,T).

aleph_member3(A,A-B):- A =< B.
aleph_member3(X,A-B):-
    A < B,
    A1 is A + 1,
    aleph_member3(X,A1-B).

aleph_member(X,[X|_]).
aleph_member(X,[_|T]):-
    aleph_member(X,T).

%----------------
goals_to_list((true,Goals),T):-
    !,
    goals_to_list(Goals,T).
goals_to_list((Goal,Goals),[Goal|T]):-
    !,
    goals_to_list(Goals,T).
goals_to_list(true,[]):- !.
goals_to_list(Goal,[Goal]).

list_to_goals([Goal],Goal):- !.
list_to_goals([Goal|Goals],(Goal,Goals1)):-
    list_to_goals(Goals,Goals1).


prune(_):-fail.

in((Head:-true),Head):- !.
in((Head:-Body),L):-
    !,
    in((Head,Body),L).
in((L1,_),L1).
in((_,R),L):-
    !,
    in(R,L).
in(L,L).

in((L1,L),L1,L).
in((L1,L),L2,(L1,Rest)):-
    !,
    in(L,L2,Rest).
in(L,L,true).

member_eq(A,[H|_T]):-
  A==H,!.

member_eq(A,[_H|T]):-
  member_eq(A,T).
